"""
Exporters for Otter Export
"""

import shutil

from .via_html import PDFViaHTMLExporter
from .via_latex import PDFViaLatexExporter

EXPORTERS = {
    "html": PDFViaHTMLExporter,
    "latex": PDFViaLatexExporter,
}

class WkhtmltopdfNotFoundError(Exception):
    """
    Exception to throw when PDF via HTML is indicated but wkhtmltopdf is not found
    """

def get_exporter(exporter_type=None):
    """
    Returns the preferred exporter to use. Defaults to PDF via LaTeX exporter. Pass a string to
    ``exporter_type`` to override this behavior.

    Args:
        exporter_type (``str``, optional): a string identifying the type of exporter to use; can be one
            of {"html", "latex"}

    Returns:
        ``otter.export.exporters.base_exporter.BaseExporter``: the exporter class

    Raises:
        ``Exception``: if PDF via HTML is indicated but wkhtmltopdf is not installed.
    """
    if exporter_type is not None:
        exporter_type = exporter_type.lower()
        assert exporter_type in EXPORTERS.keys(), f"{exporter_type} is not a valid PDF exporter"

        if exporter_type == 'html' and shutil.which("wkhtmltopdf") is None:
            raise WkhtmltopdfNotFoundError("PDF via HTML indicated but wkhtmltopdf not found")

        return EXPORTERS[exporter_type]

    return PDFViaLatexExporter



"""
Solution removal for Otter Assign
"""

import re
import nbformat

from .constants import MD_SOLUTION_REGEX
from .utils import get_source, remove_output

def is_markdown_solution_cell(cell):
    """
    Returns whether any line of the cell matches MD_SOLUTION_REGEX
    
    Args:
        cell (``nbformat.NotebookNode``): notebook cell
    
    Returns:
        ``bool``: whether the current cell is a Markdown solution cell
    """
    if not cell.cell_type == 'markdown':
        return False
    source = get_source(cell)
    return source and any([re.match(MD_SOLUTION_REGEX, l, flags=re.IGNORECASE) for l in source])

def has_seed(cell):
    if not cell.cell_type == 'code':
        return False
    source = get_source(cell)
    return source and any([l.strip().endswith('# SEED') for l in source])

solution_assignment_regex = re.compile(r"(\s*[a-zA-Z0-9_. ]*(=|<-))(.*)[ ]?#[ ]?SOLUTION")
def solution_assignment_sub(match):
    """
    """
    prefix = match.group(1)
    return prefix + ' ...'

solution_line_regex = re.compile(r"(\s*)([^#\n]+)[ ]?#[ ]?SOLUTION")
def solution_line_sub(match):
    """
    """
    prefix = match.group(1)
    return prefix + '...'

begin_solution_regex = re.compile(r"(\s*)# BEGIN SOLUTION( NO PROMPT)?")
skip_suffixes = ['# SOLUTION NO PROMPT', '# BEGIN PROMPT', '# END PROMPT', '# SEED']

SUBSTITUTIONS = [
    (solution_assignment_regex, solution_assignment_sub),
    (solution_line_regex, solution_line_sub),
]

# TODO: comments, docstrings
def replace_solutions(lines):
    """
    Replaces solutions in ``lines``
    
    Args:
        lines (``list`` of ``str``): solutions as a list of strings

    Returns:
        ``list`` of ``str``: stripped version of lines without solutions
    """
    stripped = []
    solution = False
    for line in lines:

        # ...
        if any(line.endswith(s) for s in skip_suffixes):
            continue

        # ...
        if solution and not line.endswith('# END SOLUTION'):
            continue

        # ...
        if line.endswith('# END SOLUTION'):
            assert solution, f"END SOLUTION without BEGIN SOLUTION in {lines}"
            solution = False
            continue

        # ...
        begin_solution = begin_solution_regex.match(line)
        if begin_solution:
            assert not solution, f"Nested BEGIN SOLUTION in {lines}"
            solution = True
            if not begin_solution.group(2):
                line = begin_solution.group(1) + '...'
            else:
                continue
        for exp, sub in SUBSTITUTIONS:
            m = exp.match(line)
            if m:
                line = sub(m)
        
        stripped.append(line)
    
    assert not solution, f"BEGIN SOLUTION without END SOLUTION in {lines}"
    
    return stripped

def strip_solutions_and_output(nb):
    """Write a notebook with solutions stripped
    
    Args:
        original_nb_path (path-like): path to original notebook
        stripped_nb_path (path-like): path to new stripped notebook
    """
    md_solutions = []
    for i, cell in enumerate(nb['cells']):
        cell['source'] = '\n'.join(replace_solutions(get_source(cell)))
        if is_markdown_solution_cell(cell):
            md_solutions.append(i)
    md_solutions.reverse()
    for i in md_solutions:
        del nb['cells'][i]
    
    # remove output from student version
    remove_output(nb)
    
    return nb



"""
Otter Grade command-line utility. Provides local grading of submissions in parallel Docker containers.
"""

import os
import pandas as pd

from .metadata import GradescopeParser, CanvasParser, JSONParser, YAMLParser
from .containers import launch_parallel_containers
from .utils import merge_csv

def main(args):
    """Runs Otter Grade

    Args:
        args (``argparse.Namespace``): parsed command line arguments
    """
    # Asserts that exactly one metadata flag is provided
    assert sum([meta != False for meta in [
        args.gradescope, 
        args.canvas, 
        args.json, 
        args.yaml
    ]]) <= 1, "You can specify at most one metadata flag (-g, -j, -y, -c)"

    # # Asserts that either --pdf, --tag-filter, or --html-filter but not both provided
    # assert sum([args.pdf, args.tag_filter, args.html_filter]) <= 1, "Cannot provide more than 1 PDF flag"

    # verbose flag
    verbose = args.verbose

    # Hand off metadata to parser
    if args.gradescope:
        meta_parser = GradescopeParser(args.path)
        if verbose:
            print("Found Gradescope metadata...")
    elif args.canvas:
        meta_parser = CanvasParser(args.path)
        if verbose:
            print("Found Canvas metadata...")
    elif args.json:
        meta_parser = JSONParser(os.path.join(args.json))
        if verbose:
            print("Found JSON metadata...")
    elif args.yaml:
        meta_parser = YAMLParser(os.path.join(args.yaml))
        if verbose:
            print("Found YAML metadata...")
    else:
        meta_parser = None

    # check that reqs file is valid
    requirements = args.requirements
    if requirements is None and os.path.isfile("requirements.txt"):
        requirements = "requirements.txt"
    
    if requirements:
            assert os.path.isfile(requirements), f"Requirements file {requirements} not found"

    # if not os.path.isfile(args.requirements):
        
    #     # if user-specified requirements not found, fail with AssertionError
    #     assert args.requirements == "requirements.txt", f"requirements file {args.requirements} does not exist"

    #     # else just set to None and reqs are ignored
    #     args.requirements = None

    if verbose:
        print("Launching docker containers...")

    # Docker
    grades_dfs = launch_parallel_containers(args.tests_path, 
        args.path, 
        verbose=verbose, 
        pdfs=args.pdfs,
        # unfiltered_pdfs=args.pdf, 
        # tag_filter=args.tag_filter,
        # html_filter=args.html_filter,
        # reqs=args.requirements,
        reqs=requirements,
        num_containers=args.containers,
        image=args.image,
        scripts=args.scripts,
        no_kill=args.no_kill,
        output_path=args.output_path,
        debug=args.debug,
        seed=args.seed,
        zips=args.zips,
        meta_parser=meta_parser
    )

    if verbose:
        print("Combining grades and saving...")

    # Merge Dataframes
    output_df = merge_csv(grades_dfs)

    def map_files_to_ids(row):
        """Returns the identifier for the filename in the specified row"""
        return meta_parser.file_to_id(row["file"])

    # add in identifier column
    if meta_parser is not None:
        output_df["identifier"] = output_df.apply(map_files_to_ids, axis=1)
        output_df.drop("file", axis=1, inplace=True)

        # reorder cols in output_df
        cols = output_df.columns.tolist()
        output_df = output_df[cols[-1:] + cols[:-1]]

    # write to CSV file
    output_df.to_csv(os.path.join(args.output_path, "final_grades.csv"), index=False)



"""
Otter Check command-line utility
"""

import os

from glob import glob
from jinja2 import Template

from .logs import LogEntry, EventType
from .notebook import _OTTER_LOG_FILENAME
from ..execute import grade_notebook, check
from ..utils import block_print


RESULT_TEMPLATE = Template("""{% if grade == 1.0 %}All tests passed!{% else %}{{ passed_tests|length }} of {{ tests|length }} tests passed
{% if passed_tests %}
Tests passed:
    {% for passed_test in passed_tests %}{{ passed_test }} {% endfor %}
{% endif %}
{% if failed_tests %}
Tests failed: 
{% for failed_test in failed_tests %}{{ failed_test }}{% endfor %}{% endif %}{% endif %}
""")

def _log_event(event_type, results=[], question=None, success=True, error=None):
	"""
	Logs an event

	Args:
		event_type (``otter.logs.EventType``): the type of event
		results (``list`` of ``otter.test_files.abstract_test.TestCollectionResults``, optional): the 
			results of any checks recorded by the entry
		question (``str``, optional): the question name for this check
		success (``bool``, optional): whether the operation was successful
		error (``Exception``, optional): the exception thrown by the operation, if applicable
	"""
	LogEntry(
		event_type,
		results=results,
		question=question, 
		success=success, 
		error=error
	).flush_to_file(_OTTER_LOG_FILENAME)

def main(args):
	"""
	Runs Otter Check

	Args:
		args (``argparse.Namespace``): parsed command line arguments
	"""

	try:
		if args.question:
			test_path = os.path.join(args.tests_path, args.question + ".py")
			assert os.path.isfile(test_path), "Test {} does not exist".format(args.question)
			qs = [test_path]
		else:
			qs = glob(os.path.join(args.tests_path, "*.py"))

		assert os.path.isfile(args.file), "{} is not a file".format(args.file)
		assert args.file[-6:] == ".ipynb" or args.file[-3:] == ".py", "{} is not a Jupyter Notebook or Python file".format(args.file)

		script = args.file[-3:] == ".py"

		with block_print():
			results = grade_notebook(
				args.file,
				tests_glob=qs,
				script=script,
				seed=args.seed
			)

		passed_tests = [
			results.get_result(test_name).name for test_name in results.tests if not results.get_result(test_name).incorrect
			# test for test in results if test not in ["possible", "total"] and "hint" not in results[test]
		]
		failed_tests = [
			repr(results.get_result(test_name).test) for test_name in results.tests if results.get_result(test_name).incorrect
		]

		output = RESULT_TEMPLATE.render(
			grade=results.total / results.possible,
			passed_tests=passed_tests,
			failed_tests=failed_tests,
			tests=results.tests
		)

		print(output)

	except Exception as e:
		_log_event(EventType.CHECK, success=False, error=e)
		raise e
			
	else:
		_log_event(EventType.CHECK, results=results)



################################
##### Tests for Otter Logs #####
################################

import os
import unittest

from otter.argparser import get_parser
from otter.check.logs import Log
from otter.check.notebook import Notebook, _OTTER_LOG_FILENAME
from otter.check.logs import LogEntry, EventType, Log
from . import TestCase

import numpy as np
from sklearn.linear_model import LinearRegression
import sys

parser = get_parser()

TEST_FILES_PATH = "test/test-logs/"

class TestLogs(TestCase):

    grading_results = {}
    entry_results = {}
    test_directory = TEST_FILES_PATH + "tests"

    def setUp(self):
        super().setUp()
        self.grading_results = {}
        self.entry_results = {}
        self.maxDiff = None

    def test_Notebook_check(self):
        grader = Notebook(self.test_directory)

        def square(x):
            return x**2

        for test_file in os.listdir(self.test_directory):
            if os.path.splitext(test_file)[1] != ".py":
                continue
            test_name = os.path.splitext(test_file)[0]
            self.grading_results[test_name] = grader.check(test_name)

        log = Log.from_file(_OTTER_LOG_FILENAME)

        for question in log.get_questions():
            logged_result = log.get_results(question)
            actual_result = self.grading_results[question]

            # checking repr since the results __eq__ method is not defined
            self.assertEqual(repr(logged_result), repr(actual_result), f"Logged results for {question} are not correct")


    def test_grade_check(self):
        grader = Notebook(self.test_directory)

        def square(x):
            return x**2
        
        for test_file in os.listdir(self.test_directory):
            if os.path.splitext(test_file)[1] != ".py":
                continue
            test_name = os.path.splitext(test_file)[0]
            self.grading_results[test_name] = grader.check(test_name)

        log = Log.from_file(_OTTER_LOG_FILENAME)

        for question in log.get_questions():
            logged_grade = log.get_question_entry(question).get_score_perc()
            actual_grade = self.grading_results[question].grade

            # checking repr since the results __eq__ method is not defined
            self.assertEqual(logged_grade, actual_grade, f"Logged results for {question} are not correct")


    def test_question_entry(self):
        grader = Notebook(self.test_directory)

        def square(x):
            return x**2

        for test_file in os.listdir(self.test_directory):
            if os.path.splitext(test_file)[1] != ".py":
                continue
            test_name = os.path.splitext(test_file)[0]
            self.entry_results[test_name] = grader.check(test_name)

        log = Log.from_file(_OTTER_LOG_FILENAME)

        for question in log.get_questions():
            logged_result = log.get_question_entry(question).get_results()
            actual_result = self.entry_results[question]

            # checking repr since the results __eq__ method is not defined
            self.assertEqual(repr(logged_result), repr(actual_result), f"Logged results for {question} are not correct")

    def test_shelve(self):
        # tests shelve() and unshelve() which call shelve_environment()

        def square(x):
            return x**2

        import calendar 

        env = {"num": 5, "func": square, "model": LinearRegression(), "module": sys, "ignored_func": calendar.setfirstweekday}

     
        entry = LogEntry(
            event_type=EventType.AUTH,
            results=[],
            question=None, 
            success=True, 
            error=None
        )
 

        entry.shelve(env, delete=True, filename = _OTTER_LOG_FILENAME, ignore_modules=['calendar'])
        self.assertTrue(entry.shelf)
        self.assertEqual(entry.unshelved, ["module", "ignored_func"])

        entry.flush_to_file(_OTTER_LOG_FILENAME) 

        from math import factorial

        log = Log.from_file(_OTTER_LOG_FILENAME)
        env = entry.unshelve()
        self.assertEqual([*env], ["num", "func", "model"])

        env_with_factorial = entry.unshelve(dict(factorial = factorial ))
        self.assertTrue("factorial" in env_with_factorial["func"].__globals__)
        self.assertTrue(factorial is env_with_factorial["func"].__globals__["factorial"])


    
    
    def test_Log_getItem(self):
        entry1 = LogEntry(
            event_type=EventType.AUTH,
            results=[],
            question=None, 
            success=True, 
            error=None
        )

        entry1.flush_to_file(_OTTER_LOG_FILENAME)

        log = Log.from_file(_OTTER_LOG_FILENAME)

        self.assertEqual(log.__getitem__(0).event_type, entry1.event_type)
        

    def test_Log_iter(self):
        #log = Log.from_file(_OTTER_LOG_FILENAME, ascending = True)
        #list of entries
        #logIter = log.iter()

        entry1 = LogEntry(
            event_type=EventType.CHECK,
            results=[],
            question= "q1", 
            success=True, 
            error=None
        )

        entry2 = LogEntry(
            event_type=EventType.CHECK,
            results=[],
            question= "q1", 
            success=True, 
            error=None
        )

        entry3 = LogEntry(
            event_type=EventType.CHECK,
            results=[],
            question= "q2", 
            success=True, 
            error=None
        )

        entry1.flush_to_file(_OTTER_LOG_FILENAME)
        entry2.flush_to_file(_OTTER_LOG_FILENAME)
        entry3.flush_to_file(_OTTER_LOG_FILENAME)

        log = Log.from_file(_OTTER_LOG_FILENAME)

        logIter = log.question_iterator() #most recent entries / question
        self.assertEqual(logIter.questions , ["q1", "q2"])

        nextLogEntry = next(logIter)

        self.assertEqual(nextLogEntry.question, entry2.question)

        nextLogEntry = next(logIter)

        self.assertEqual(nextLogEntry.question, entry3.question)

    def tearDown(self):
        if os.path.isfile(_OTTER_LOG_FILENAME):
            os.remove(_OTTER_LOG_FILENAME)



"""
Execution of a submission through log deserialization
"""

import re

from unittest import mock
from IPython.core.inputsplitter import IPythonInputSplitter

from ..utils import hide_outputs

def execute_log(nb, log, secret='secret', initial_env=None, ignore_errors=False, cwd=None, test_dir=None, variables=None):
    """
    Executes a notebook from logged environments and returns the global environment that results from execution

    Execute notebook & return the global environment that results from execution. If ``ignore_errors`` 
    is ``True``, exceptions are swallowed. ``secret`` contains random digits so ``check_results`` and 
    ``check`` are not easily modifiable. ``nb`` is passed in as a dictionary that's a parsed notebook

    Args:
        nb (``dict``): JSON representation of a notebook
        log (``otter.check.logs.Log``): log from notebook execution
        secret (``str``, optional): randomly generated integer used to rebind check function
        initial_env (``str``, optional): name of initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        variables (``dict``, optional): map of variable names -> type string to check type of deserialized
            object to prevent arbitrary code from being put into the environment
    
    Results:
        ``dict``: global environment resulting from executing all code of the input notebook
    """
    with hide_outputs():
        if initial_env:
            global_env = initial_env.copy()
        else:
            global_env = {}

        if test_dir:
            source = f"import otter\ngrader = otter.Notebook(\"{test_dir}\")\n"
        else:
            source = f"import otter\ngrader = otter.Notebook()\n"
        
        if cwd:
            source +=  f"import sys\nsys.path.append(\"{cwd}\")\n"

        logged_questions = []
        m = mock.mock_open()
        with mock.patch("otter.Notebook._log_event", m):
            exec(source, global_env)

            for cell in nb['cells']:
                if cell['cell_type'] == 'code':
                    # transform the input to executable Python
                    # FIXME: use appropriate IPython functions here
                    isp = IPythonInputSplitter(line_input_checker=False)
                    
                    code_lines = []
                    cell_source_lines = cell['source']
                    source_is_str_bool = False
                    if isinstance(cell_source_lines, str):
                        source_is_str_bool = True
                        cell_source_lines = cell_source_lines.split('\n')

                    # only execute import statements
                    cell_source_lines = [re.sub(r"^\s+", "", l) for l in cell_source_lines if "import" in l]                                
                    
                    for line in cell_source_lines:
                        try:
                            exec(line, global_env)
                    # source += cell_source
                        except:
                            if not ignore_errors:
                                raise


            for entry in log.question_iterator():
                shelf = entry.unshelve(global_env)

                if variables is not None:
                    for k, v in shelf.items():
                        full_type = type(v).__module__ + "." + type(v).__name__
                        if not (k in variables and variables[k] == full_type):
                            del shelf[k]
                            print(f"Found variable of different type than expected: {k}")

                global_env.update(shelf)
                global_env[f"check_results_{secret}"].append(global_env["grader"].check(entry.question, global_env=global_env))
                logged_questions.append(entry.question)

        print("Questions executed from log: {}".format(", ".join(logged_questions)))
        
        return global_env



####################################
##### Tests for otter generate #####
####################################
import os
import unittest
import subprocess
import json
import shutil

from subprocess import PIPE
from glob import glob
from unittest import mock
from shutil import copyfile

from otter.argparser import get_parser
from otter.generate.autograder import main as autograder
from otter.generate.run_autograder import main as run_autograder

from .. import TestCase

parser = get_parser()

TEST_FILES_PATH = "test/test_generate/test-run-autograder/"

class TestRunAutograder(TestCase):
    def setUp(self):
        super().setUp()

        self.env = {}
        with open(TEST_FILES_PATH + "run_autograder") as f:
            exec(f.read(), self.env)
        
        self.config = self.env["config"]

    def test_run_autograder(self):

        #generate the zip file 
        generate_command = ["generate", "autograder",
            "-t", TEST_FILES_PATH + "tests",
            "-o", TEST_FILES_PATH,
            "-r", TEST_FILES_PATH + "requirements.txt",
            TEST_FILES_PATH + "data/test-df.csv"
        ]
        args = parser.parse_args(generate_command)
        args.func = autograder
        args.func(args)

        # first unzip and check output
        os.mkdir(TEST_FILES_PATH + "autograder")
        unzip_command = ["unzip", "-o", TEST_FILES_PATH + "autograder.zip", "-d", TEST_FILES_PATH + "autograder/source"]
        unzip = subprocess.run(unzip_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(unzip.stderr), 0, unzip.stderr.decode("utf-8"))

        self.config["autograder_dir"] = TEST_FILES_PATH + "autograder"

        # copy submission tests and notebook, 
        os.mkdir(TEST_FILES_PATH + "autograder/submission")
        os.mkdir(TEST_FILES_PATH + "autograder/results")
        copyfile(TEST_FILES_PATH + "fails2and6H.ipynb", TEST_FILES_PATH + "autograder/submission/fails2and6H.ipynb")

        run_autograder(self.config)

        self.assertDirsEqual(TEST_FILES_PATH + "autograder", TEST_FILES_PATH + "autograder-correct", ignore_ext=[".pdf",".zip"], ignore_dirs=["__pycache__"])

    def tearDown(self):
        for p in [TEST_FILES_PATH + "autograder", TEST_FILES_PATH + "autograder.zip", "test/results.json", TEST_FILES_PATH + "autograder-correct/__pycache__"]:
            if os.path.isdir(p):
                shutil.rmtree(p)
            elif os.path.isfile(p):
                os.remove(p)



"""
Utilities for Otter Generate
"""

import re
import urllib
import nbformat

from ..utils import get_source

def requirements_to_conda_env(requirements):
    """
    Converts a requirements.txt file string to a Pythonic environment.yml format

    Args:
        requirements (``str``): the requirements.txt contents
    
    Returns:
        ``dict``: the environment.yml
    """
    requirements = [l for l in requirements.split("\n") if l.strip() and not l.startswith("#")]
    conda_env = {
        "dependencies": {
            "pip": requirements
        }
    }
    return conda_env

def replace_notebook_instances(nb_path):
    """
    Replaces the creation of ``otter.Notebook`` instances in a notebook at ``nb_path`` that have custom 
    test paths with the default for Gradescope.

    Args:
        nb_path (``str``): path to the notebook
    """
    nb = nbformat.read(nb_path, as_version=nbformat.NO_CONVERT)

    instance_regex = r"otter.Notebook\([\"'].+[\"']\)"
    for cell in nb['cells']:
        source = get_source(cell)
        for i, line in enumerate(source):
            line = re.sub(instance_regex, "otter.Notebook()", line)
            source[i] = line
        cell['source'] = "\n".join(source)

    nbformat.write(nb, nb_path)



"""
Output writing for Otter Assign
"""

import os
import shutil
import pathlib
import warnings
import nbformat

from .constants import NB_VERSION
from .notebook_transformer import transform_notebook
from .solutions import strip_solutions_and_output
from .tests import write_test

def write_autograder_dir(nb_path, output_nb_path, assignment, args):
    """
    Converts a master notebook to a solutions notebook and writes this notebook to the output directory,
    copying support files and writing tests as needed.

    Args:
        nb_path (``pathlib.Path``): path to master notebook
        output_nb_path (``pathlib.Path``): path to output file
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        args (``argparse.Namespace``): parsed command line arguments
    """
    with open(nb_path) as f:
        nb = nbformat.read(f, as_version=NB_VERSION)

    if assignment.lang is None:
        try:
            lang = nb["metadata"]["kernelspec"]["language"].lower()
            assignment.lang = lang
        except IndexError:
            warnings.warn("Could not auto-parse kernelspec from notebook; assuming Python")
            assignment.lang = "python"

    output_dir = output_nb_path.parent
    tests_dir = output_dir / 'tests'
    os.makedirs(tests_dir, exist_ok=True)

    if assignment.requirements:
        shutil.copy(assignment.requirements, str(output_dir / 'requirements.txt'))

    transformed_nb, test_files = transform_notebook(nb, assignment, args)

    # write notebook
    # with open(output_nb_path) as f:
    # nbformat.write(transformed_nb, )
    nbformat.write(transformed_nb, str(output_nb_path))

    # write tests
    test_ext = (".R", ".py")[assignment.is_python]
    for test_name, test_file in test_files.items():
        write_test(tests_dir / (test_name + test_ext), test_file)

    # copy files
    for file in assignment.files or args.files:

        # if a directory, copy the entire dir
        if os.path.isdir(file):
            shutil.copytree(file, str(output_dir / os.path.basename(file)))
            
        else:
            # check that file is in subdir
            assert os.path.abspath(nb_path.parent) in os.path.abspath(file), \
                f"{file} is not in a subdirectory of the master notebook directory"
            file_path = pathlib.Path(file)
            rel_path = file_path.parent.relative_to(nb_path.parent)
            os.makedirs(output_dir / rel_path, exist_ok=True)
            shutil.copy(file, str(output_dir / rel_path))

def write_student_dir(nb_name, autograder_dir, student_dir, assignment, args):
    """
    Copies the autograder (solutions) directory and removes extraneous files, strips solutions from
    the notebook, and removes hidden tests from the tests directory.

    Args:
        nb_name (``str``): the master notebook name
        autograder_dir (``pathlib.Path``): the path to the autograder directory
        student_dir (``pathlib.Path``): the path to the student directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        args (``argparse.Namespace``): parsed command line arguments
    """
    if assignment.is_r:
        from .r_adapter.tests import remove_hidden_tests_from_dir
    else:
        from .tests import remove_hidden_tests_from_dir

    # copy autograder dir
    shutil.copytree(autograder_dir, student_dir)

    # remove requirements from student dir if present
    requirements = str(student_dir / 'requirements.txt')
    if os.path.isfile(requirements):
        os.remove(requirements)

    # strip solutions from student version
    student_nb_path = student_dir / nb_name
    with open(student_nb_path) as f:
        nb = nbformat.read(f, as_version=NB_VERSION)

    nb = strip_solutions_and_output(nb)

    with open(student_nb_path, "w") as f:
        nbformat.write(nb, f)

    # remove hidden tests from student directory
    remove_hidden_tests_from_dir(student_dir / 'tests', assignment)

def write_output_directories(master_nb_path, result_dir, assignment, args):
    """
    Converts a master notebook to an autograder and student directory based on configurations in 
    ``assignment`` and ``args``.

    Args:
        master_nb_path (``nbformat.NotebookNode``): the master notebook path
        result_dir (``pathlib.Path``): path to the result directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        args (``argparse.Namespace``): parsed command line arguments
    """
    # create directories
    autograder_dir = result_dir / 'autograder'
    student_dir = result_dir / 'student'
    shutil.rmtree(autograder_dir, ignore_errors=True)
    shutil.rmtree(student_dir, ignore_errors=True)
    os.makedirs(autograder_dir, exist_ok=True)

    # write autograder directory
    output_nb_path = autograder_dir / master_nb_path.name
    write_autograder_dir(master_nb_path, output_nb_path, assignment, args)

    # write student dir
    write_student_dir(master_nb_path.name, autograder_dir, student_dir, assignment, args)



"""
Grading results from submission execution
"""

import os
import math
import pprint

from collections import namedtuple

from ..check.logs import QuestionNotInLogException
from ..generate.constants import DEFAULT_OPTIONS

TestResult = namedtuple("TestResult", ["name", "score", "possible", "test", "hidden", "incorrect"])

class GradingResults:
    """
    Stores and wrangles test result objects.
    
    Initialize with a list of ``otter.test_files.abstract_test.TestCollectionResults`` objects and 
    this class will store the results as named tuples so that they can be accessed/manipulated easily. 
    Also contains methods to put the results into a nice ``dict`` format or into the correct format 
    for Gradescope.

    Args:
        results (``list`` of ``otter.test_files.abstract_test.TestCollectionResults``): the list of 
            grading results
    
    Attributes:
        raw_results (``list`` of ``otter.test_files.abstract_test.TestCollectionResults``): the 
            results passed to the constructor
        results (``dict``): maps test names to ``TestResult`` named tuples containing the test result
            information
        total (numeric): the total points earned by the submission
        possible (numeric): the total points possible based on the tests
        tests (``list`` of ``str``): list of test names according to the keys of ``results``
    """
    def __init__(self, results):
        self.raw_results = results
        self.results = {}
        
        total_score, points_possible = 0, 0
        for result in results:
            for test in result.tests:
                test_name = os.path.splitext(os.path.basename(test.name))[0]
                tr = TestResult(
                    name = test_name,
                    score = result.grade * test.value,
                    possible = test.value,
                    test = test,
                    hidden = False,
                    incorrect = False
                )
                self.results[test_name] = tr

                total_score += result.grade * test.value
                points_possible += test.value

            for test, test_obj in result.failed_tests:
                test_name = os.path.splitext(os.path.basename(test.name))[0]
                if test_name in self.results:
                    self.results[test_name] = self.results[test_name]._replace(
                        hidden = test_obj.failed_test_hidden,
                        incorrect = True
                    )
                else:
                    self.results[test_name] = TestResult(
                        name = test_name,
                        score = 0,
                        possible = test.value,
                        test = test,
                        hidden = test_obj.failed_test_hidden,
                        incorrect = True
                    )
        
        self.total = total_score
        self.possible = points_possible
    
    def __repr__(self):
        return pprint.pformat(self.to_dict(), indent=2)

    @property
    def tests(self):
        return list(self.results.keys())
    
    def get_result(self, test_name):
        """
        Returns the ``TestResult`` named tuple corresponding to the test with name ``test_name``

        Args:
            test_name (``str``): the name of the desired test
        
        Returns:
            ``TestResult``: the results of that test
        """
        return self.results[test_name]

    def get_score(self, test_name):
        """
        Returns the score of a test tracked by these results

        Args:
            test_name (``str``): the name of the test
        
        Returns:
            ``int`` or ``float``: the score
        """
        result = self.results[test_name]
        return result.score

    def get_public_score(self, test_name):
        """
        Returns the score of a question based on only public tests. Assumes that all public tests in
        a test file occur before the first hidden tests (because test execution stops at the first
        failed test).

        Args:
            test_name (``str``): the name of the test
        
        Returns:
            ``int`` or ``float``: the score based only on public tests
        """
        result = self.results[test_name]
        if not result.incorrect:
            return result.possible
        elif result.hidden:
            return result.possible
        else:
            return result.score

    def verify_against_log(self, log, ignore_hidden=True):
        """
        Verifies these scores against the results stored in this log using the results returned by 
        ``Log.get_results`` for comparison. Prints a message if the scores differ by more than the 
        default tolerance of ``math.isclose``. If ``ignore_hidden`` is ``True``, hidden tests are
        ignored when verifying scores.

        Args:
            log (``otter.logs.Log``): the log to verify against
            ignore_hidden  (``bool``, optional): whether to ignore hidden tests during verification

        Returns:
            ``bool``: whether a discrepancy was found
        """
        found_discrepancy = False
        for test_name in  self.tests:
            if ignore_hidden:
                score = self.get_public_score(test_name)
            else:
                score = self.get_score(test_name)
            try:
                result = log.get_results(test_name)
                logged_score = result.grade * result.tests[0].value
                if not math.isclose(score, logged_score):
                    print("Score for {} ({:.3f}) differs from logged score ({:.3f})".format(
                        test_name, score, logged_score
                    ))
                    found_discrepancy = True
            except QuestionNotInLogException:
                print(f"No score for {test_name} found in this log")
                found_discrepancy = True
        return found_discrepancy

    def to_dict(self):
        """
        Converts these results into a dictinary, extending the fields of the named tuples in ``results``
        into key, value pairs in a ``dict``.

        Returns:
            ``dict``: the results in dictionary form
        """
        output = {}
        for test_name in self.tests:
            result = self.get_result(test_name)
            output[test_name] = dict(result._asdict())

        return output

    def to_gradescope_dict(self, config={}):
        """
        Converts these results into a dictionary formatted for Gradescope's autograder. Requires a 
        dictionary of configurations for the Gradescope assignment generated using Otter Generate.

        Args:
            config (``dict``): the grading configurations

        Returns:
            ``dict``: the results formatted for Gradescope
        """
        options = DEFAULT_OPTIONS.copy()
        options.update(config)

        output = {"tests": []}

        # hidden visibility determined by show_hidden_tests_on_release
        hidden_test_visibility = ("hidden", "after_published")[options["show_hidden_tests_on_release"]]
        no_separate_visibility = options["test_visibility"]
        assert no_separate_visibility in ["hidden", "visible", "after_published"]

        for test_name in self.tests:
            result = self.get_result(test_name)
            hidden, incorrect = result.hidden, result.incorrect
            score, possible = result.score, result.possible
            public_score, hidden_score = score * options["public_multiplier"], score * (1 - options["public_multiplier"])
            public_possible, hidden_possible = possible * options["public_multiplier"], possible * (1 - options["public_multiplier"])
        
            if hidden and incorrect:
                public_score, hidden_score = possible * options["public_multiplier"], 0
            elif not hidden and incorrect:
                public_score, hidden_score = 0, 0
                public_possible = possible
            
            if options["separate_tests"]:
                output["tests"].append({
                    "name": result.name + " - Public",
                    "score": public_score,
                    "max_score": public_possible,
                    "visibility": "visible",
                    "output": repr(result.test) if not hidden and incorrect else "All tests passed!"
                })

                if not (not hidden and incorrect):
                    output["tests"].append({
                        "name" : result.name + " - Hidden",
                        "score" : hidden_score,
                        "max_score": hidden_possible,
                        "visibility": hidden_test_visibility,
                        "output": repr(result.test) if incorrect else "All tests passed!"
                    })
            
            else:
                output["tests"].append({
                    "name": result.name,
                    "score": score,
                    "max_score": possible,
                    "visibility": no_separate_visibility,
                    "output": repr(result.test) if not hidden and incorrect else "All tests passed!"
                })
        
        if options["show_stdout_on_release"]:
            output["stdout_visibility"] = "after_published"

        if options["points_possible"] is not None:
            output["score"] = self.total / self.possible * options["points_possible"]

        if options["score_threshold"] is not None:
            if self.total / self.possible >= config["score_threshold"]:
                output["score"] = options["points_possible"] or self.possible
            else:
                output["score"] = 0
        
        return output



"""
PDF via LaTeX exporter
"""

import os
import warnings
import pkg_resources
import nbconvert

from .base_exporter import BaseExporter

class PDFViaLatexExporter(BaseExporter):
    """
    Exports notebooks to PDF files using LaTeX as an intermediary

    Converts IPython notebooks to PDFs by first converting them into temporary TeX files that are then
    converted to PDFs using nbconvert and pandoc. Pagebreaks, if enabled, are enforced with a custom
    LaTeX template that clears the document to the next odd numbered page, resulting in responses that
    are all two pages long.

    Attributes:
        default_options (``dict``): the default options for this exporter
    """

    default_options = BaseExporter.default_options.copy()
    default_options.update({
        "save_tex": False,
        "template": "templates/via_latex.tpl",
    })

    @classmethod
    def convert_notebook(cls, nb_path, dest, debug=False, **kwargs):
        warnings.filterwarnings("ignore", r"invalid escape sequence '\\c'", DeprecationWarning)

        options = cls.default_options.copy()
        options.update(kwargs)

        nb = cls.load_notebook(nb_path, filtering=options["filtering"], pagebreaks=options["pagebreaks"])

        if options["save_tex"]:
            latex_exporter = nbconvert.LatexExporter()
            latex_exporter.template_file = pkg_resources.resource_filename(__name__, options["template"])

        pdf_exporter = nbconvert.PDFExporter()
        pdf_exporter.template_file = pkg_resources.resource_filename(__name__, options["template"])

        try:
            pdf_output = pdf_exporter.from_notebook_node(nb)
            with open(dest, "wb") as output_file:
                output_file.write(pdf_output[0])

            if options["save_tex"]:
                latex_output = latex_exporter.from_notebook_node(nb)
                with open(os.path.splitext(dest)[0] + ".tex", "w+") as output_file:
                    output_file.write(latex_output[0])

        except nbconvert.pdf.LatexFailed as error:
            print("There was an error generating your LaTeX")
            output = error.output
            if debug:
                print("Showing full error message from PDFTex")

            else:
                print("Showing concise error message")
                output = "\n".join(error.output.split("\n")[-15:])

            print("=" * 60)
            print(output)
            print("=" * 60)



"""
In-container grading interface for Otter Grader
"""

import argparse
import os
import re
import json
import zipfile
import shutil
import tempfile
import pandas as pd

from glob import glob
try:
    from IPython.core.inputsplitter import IPythonInputSplitter
except ImportError:
    raise ImportError('IPython needs to be installed for notebook grading')

from ..execute import grade_notebook
from ..export import export_notebook
from ..utils import id_generator

def grade(ipynb_path, pdf, script, ignore_errors=True, seed=None, cwd=None):
    """
    Grades a single ipython notebook and returns the score

    If no PDF is needed, set the pdf, tag_filter, and html_filter parameters to false. For .py
    files, set script to true.

    Args:
        ipynb_path (``str``): path to the notebook
        pdf (``bool``): whether unfiltered PDFs of notebooks should be generated
        tag_filter (``bool``): whether cell tag-filtered PDFs of notebooks should be generated
        html_filter (``bool``): whether HTML comment-filtered PDFs of notebooks should be generated
        script (``bool``): whether the input file is a Python script
        ignore_errors (``bool``, optional): whether errors should be ignored during execution
        seed (``int``, optional): random seed for intercell seeding
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment

    Returns:
        ``dict``: a score mapping with keys for each test, the student's scores, and total points 
            earned and possible 
    """
    # # get path of notebook file
    # base_path = os.path.dirname(ipynb_path)

    # glob tests
    test_files = glob('/home/tests/*.py')

    # get score
    result = grade_notebook(ipynb_path, test_files, script=script, ignore_errors=ignore_errors, seed=seed, cwd=cwd)

    # output PDF
    if pdf:
        export_notebook(
            ipynb_path, 
            filtering = pdf != "unfiltered", 
            debug=True,
        )

    return result

def main(args=None):
    """
    Parses command line arguments and executes submissions. Writes grades to a CSV file and optionally
    generates PDFs of submissions.

    Args:
        args (``list`` of ``str``, optional): alternate command line arguments
    """
    # implement argparser
    parser = argparse.ArgumentParser()
    parser.add_argument('submission_directory', help="Path to submissions directory")
    parser.add_argument("--pdfs", default=False, const="unfiltered", choices=["unfiltered", "html"], nargs="?")
    parser.add_argument("--scripts", action="store_true", default=False)
    parser.add_argument("--seed", type=int, default=None, help="A random seed to be executed before each cell")
    parser.add_argument("--zips", action="store_true", default=False, help="Whether the submissions are Notebook.export zip archives")
    parser.add_argument("--verbose", default=False, action="store_true", help="If present prints scores and hints to stdout")
    parser.add_argument("--debug", default=False, action="store_true", help="Does not ignore errors on execution")

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    # cd into dir containing submissions
    subs_dir = os.path.abspath(args.submission_directory)
    os.chdir(subs_dir)

    # get file extension and get all of those files in dir_path
    file_extension = (".zip", ".py", ".ipynb")[[args.zips, args.scripts, True].index(True)]
    files = [(f, os.path.join(subs_dir, f)) for f in os.listdir(subs_dir) if os.path.splitext(f)[1] == file_extension]

    all_results = {"file": [], "score": [], "manual": []}

    if not args.pdfs:
        del all_results["manual"]

    # if zips, creat subdir to grade in
    if file_extension == ".zip":
        secret = id_generator()
        grading_dir = os.path.join(subs_dir, f"SUBMISSION_GRADING_DIRECTORY_{secret}")

        # copy all support files/folders into a template grading dir which we will copy from 
        # each time we grade a submission
        template_grading_dir = tempfile.mkdtemp()
        for fn in os.listdir(subs_dir):
            fp = os.path.join(subs_dir, fn)
            if os.path.isfile(fp) and os.path.splitext(fp)[1] != file_extension:
                shutil.copy(fp, template_grading_dir)
            elif os.path.isdir(fp):
                shutil.copytree(fp, os.path.join(template_grading_dir, fn))
    
    else:
        grading_dir = None

    # grade files
    for fname, fpath in files:
        os.chdir(subs_dir)

        all_results["file"].append(fname)

        # if zips, extract zip file into grading dir and find the gradeable filename
        if file_extension == ".zip":
            shutil.copytree(template_grading_dir, grading_dir)
            zf = zipfile.ZipFile(fpath)
            zf.extractall(grading_dir)
            os.chdir(grading_dir)
            gradeables = glob(os.path.join(grading_dir, ("*.py", "*.ipynb")[not args.scripts]))
            assert len(gradeables) == 1, f"Wrong number of gradeable files in submission {fname}"
            fpath = gradeables[0]

        # grade the submission
        score = grade(
            fpath, 
            args.pdfs, 
            args.scripts, 
            ignore_errors = not args.debug, 
            seed = args.seed, 
            cwd = grading_dir if grading_dir else subs_dir
        )
        score = score.to_dict()

        if args.verbose:
            print("Score details for {}".format(fname))
            print(json.dumps(score, default=lambda o: repr(o), indent=2))

        all_results["score"].append({t : score[t]["score"] if type(score[t]) == dict else score[t] for t in score})
        
        if args.pdfs:
            pdf_path = os.path.splitext(fpath)[0] + ".pdf"
            all_results["manual"].append(pdf_path)

        if file_extension == ".zip":
            shutil.rmtree(grading_dir)

    os.chdir(subs_dir)

    if file_extension ==  ".zip":
        shutil.rmtree(template_grading_dir)

    try:
        # expand mappings in all_results["score"]
        for q in all_results["score"][0].keys():
            all_results[q] = []

        for score in all_results["score"]:
            for q in score:
                all_results[q] += [score[q]]#["score"]]

    except IndexError:
        pass

    del all_results["score"]

    final_results = pd.DataFrame(all_results)
    final_results.to_csv("grades.csv", index=False)

if __name__ == "__main__":
    main()



"""
IPython notebook PDF Exporter forked from nb2pdf and gsExport
"""

import os
import warnings
import nbformat
import nbconvert
import pkg_resources

from .exporters import get_exporter

def export_notebook(nb_path, dest=None, debug=False, exporter_type=None, **kwargs):
    """
    Exports a notebook file at ``nb_path`` to a PDF with optional filtering and pagebreaks. Accepts
    other ``kwargs`` passed to the exporter class's ``convert_notebook`` class method.

    Args:
        nb_path (``str``): path to notebook
        dest (``str``, optional): path to write PDF
        debug (``bool``, optional): whether to run export in debug mode
        exporter_type (``str``, optional): the type of exporter to use; one of {'html', 'latex'}
    """
    # notebook = load_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)

    if dest is not None:
        pdf_name = dest
    else:
        pdf_name = os.path.splitext(nb_path)[0] + ".pdf"
        
    # notebook_to_pdf(notebook, pdf_name, save_tex=save_tex, debug=debug)
    Exporter = get_exporter(exporter_type=exporter_type)
    Exporter.convert_notebook(nb_path, pdf_name, debug=debug, **kwargs)

def main(args):
    """
    Runs Otter Export

    Args:
        args (``argparse.Namespace``): parsed command line arguments
    """
    export_notebook(
        args.source,
        dest = args.dest,
        exporter_type = args.exporter,
        filtering = args.filtering,
        pagebreaks = args.pagebreaks,
        save_tex = args.save,
        save_html = args.save,
        debug = args.debug
    )



"""
Gradescope autograder configuration generator for Otter Generate
"""

import os
import shutil
import subprocess
import pathlib
import pkg_resources

from glob import glob
from subprocess import PIPE
from jinja2 import Template

from .token import APIClient

TEMPLATES_DIR = pkg_resources.resource_filename(__name__, "templates")
MINICONDA_INSTALL_URL = "https://repo.anaconda.com/miniconda/Miniconda3-py37_4.8.3-Linux-x86_64.sh"
OTTER_ENV_NAME = "otter-gradescope-env"
TEMPLATE_FILE_PATHS = {
    "setup.sh": os.path.join(TEMPLATES_DIR, "setup.sh"),
    "requirements.txt": os.path.join(TEMPLATES_DIR, "requirements.txt"),
    "requirements.r": os.path.join(TEMPLATES_DIR, "requirements.r"),
    "run_autograder": os.path.join(TEMPLATES_DIR, "run_autograder"),
    "run_otter.py": os.path.join(TEMPLATES_DIR, "run_otter.py"),
    "environment.yml":  os.path.join(TEMPLATES_DIR, "environment.yml"),
}

def main(args):
    """
    Runs ``otter generate autograder``
    """
    assert args.threshold is None or 0 <= args.threshold <= 1, "{} is not a valid threshold".format(
        args.threshold
    )

    if args.course_id or args.assignment_id:
        assert args.course_id and args.assignment_id, "Either course ID or assignment ID unspecified for PDF submissions"
        if not args.token:
            args.token = APIClient.get_token()

    # check that args.public_multiplier is valid
    assert 0 <= args.public_multiplier <= 1, f"Public test multiplier {args.public_multiplier} is not between 0 and 1"

    # check for valid language
    args.lang = args.lang.lower()
    assert args.lang.lower() in ["python", "r"], f"{args.lang} is not a supported language"

    templates = {}
    for fn, fp in TEMPLATE_FILE_PATHS.items():
        with open(fp) as f:
            templates[fn] = Template(f.read())

    # format run_autograder
    run_otter_py = templates["run_otter.py"].render(
        threshold = str(args.threshold),
        points = str(args.points),
        show_stdout = str(args.show_stdout),
        show_hidden = str(args.show_hidden),
        seed = str(args.seed),
        token = str(args.token),
        course_id = str(args.course_id),
        assignment_id = str(args.assignment_id),
        filtering = str(not args.unfiltered_pdfs),
        pagebreaks = str(not args.no_pagebreaks),
        grade_from_log = str(args.grade_from_log),
        serialized_variables = str(args.serialized_variables),
        public_multiplier = str(args.public_multiplier),
        lang = str(args.lang.lower()),
        autograder_dir = str(args.autograder_dir),
    )

    run_autograder = templates["run_autograder"].render(
        autograder_dir = str(args.autograder_dir),
    )

    # format setup.sh
    setup_sh = templates["setup.sh"].render(
        autograder_dir = str(args.autograder_dir),
        miniconda_install_url = MINICONDA_INSTALL_URL,
        ottr_branch = "stable",
        otter_env_name = OTTER_ENV_NAME,
    )

    environment_yml = templates["environment.yml"].render(
        otter_env_name = OTTER_ENV_NAME,
    )

    # create tmp directory to zip inside
    os.mkdir("./tmp")

    try:
        # copy tests into tmp
        os.mkdir(os.path.join("tmp", "tests"))
        pattern = ("*.py", "*.[Rr]")[args.lang.lower() == "r"]
        for file in glob(os.path.join(args.tests_path, pattern)):
            shutil.copy(file, os.path.join("tmp", "tests"))

        # open requirements if it exists
        requirements = args.requirements
        reqs_filename = f"requirements.{'R' if args.lang.lower() == 'r' else 'txt'}"
        if requirements is None and os.path.isfile(reqs_filename):
            requirements = reqs_filename
        
        if requirements:
            assert os.path.isfile(requirements), f"Requirements file {requirements} not found"
            f = open(requirements)
        else:
            f = open(os.devnull)

        # render the templates
        python_requirements = templates["requirements.txt"].render(
            other_requirements = f.read() if args.lang.lower() == "python" else "",
            overwrite_requirements = args.lang.lower() == "python" and args.overwrite_requirements
        )

        # reset the stream
        f.seek(0)

        r_requirements = templates["requirements.r"].render(
            other_requirements = f.read() if args.lang.lower() == "r" else "",
            overwrite_requirements = args.lang.lower() == "python" and args.overwrite_requirements

        )

        # close the stream
        f.close()
        
        # copy requirements into tmp
        with open(os.path.join(os.getcwd(), "tmp", "requirements.txt"), "w+") as f:
            f.write(python_requirements)

        with open(os.path.join(os.getcwd(), "tmp", "requirements.r"), "w+") as f:
            f.write(r_requirements)

        if r_requirements:
            with open(os.path.join(os.getcwd(), "tmp", "requirements.r"), "w+") as f:
                f.write(r_requirements)

        # write setup.sh and run_autograder to tmp
        with open(os.path.join(os.getcwd(), "tmp", "setup.sh"), "w+") as f:
            f.write(setup_sh)

        with open(os.path.join(os.getcwd(), "tmp", "run_autograder"), "w+") as f:
            f.write(run_autograder)

        with open(os.path.join(os.getcwd(), "tmp", "run_otter.py"), "w+") as f:
            f.write(run_otter_py)

        with open(os.path.join(os.getcwd(), "tmp", "environment.yml"), "w+") as f:
            f.write(environment_yml)

        # copy files into tmp
        if len(args.files) > 0:
            os.mkdir(os.path.join("tmp", "files"))

            for file in args.files:
                # if a directory, copy the entire dir
                if os.path.isdir(file):
                    shutil.copytree(file, str(os.path.join("tmp", "files", os.path.basename(file))))
                else:
                    # check that file is in subdir
                    file = os.path.abspath(file)
                    assert os.getcwd() in file, \
                        f"{file} is not in a subdirectory of the working directory"
                    wd_path = pathlib.Path(os.getcwd())
                    file_path = pathlib.Path(file)
                    rel_path = file_path.parent.relative_to(wd_path)
                    output_path = os.path.join("tmp", "files", rel_path)
                    os.makedirs(output_path, exist_ok=True)
                    shutil.copy(file, output_path)

        os.chdir("./tmp")

        zip_path = os.path.join("..", args.output_path, "autograder.zip")
        if os.path.exists(zip_path):
            os.remove(zip_path)

        zip_cmd = ["zip", "-r", zip_path, "run_autograder", "run_otter.py", "requirements.r",
                "setup.sh", "requirements.txt", "environment.yml", "tests"]
        
        if r_requirements:
            zip_cmd += ["requirements.r"]

        if args.files:
            zip_cmd += ["files"]

        zipped = subprocess.run(zip_cmd, stdout=PIPE, stderr=PIPE)

        if zipped.stderr.decode("utf-8"):
            raise Exception(zipped.stderr.decode("utf-8"))

        # move back to tmp parent directory
        os.chdir("..")
    
    except:
        # delete tmp directory
        shutil.rmtree("tmp")

        # raise the exception
        raise
    
    else:
        # delete tmp directory
        shutil.rmtree("tmp")



####################################
##### Tests for otter generate #####
####################################
import os
import unittest
import subprocess
import json
import shutil
import requests
import getpass

from subprocess import PIPE
from glob import glob
from unittest.mock import patch, mock_open
from shutil import copyfile

from otter.argparser import get_parser
from otter.generate.autograder import main as autograder
from otter.generate.run_autograder import main as run_autograder
from otter.generate.autograder import main as autograder
from otter.generate.token import APIClient

from .. import TestCase

parser = get_parser()

client = APIClient('token1')

# mock input
def get_input(text):
    return input(text)

# This method will be used by the mock to replace requests.post
def mocked_requests_post(*args, **kwargs):
    class MockResponse:
        def __init__(self, json_data, status_code):
            self.text = json_data
            self.status_code = status_code

        def json(self):
            return self.text

    return MockResponse({"token": "value1"}, 200)

class TestToken(TestCase):

    @patch.object(getpass,"getpass")
    @patch.object(APIClient,"post")
    @patch('builtins.input', return_value='user')
    def test_get_token(self, mockGetPass, mockPost, mockInput):

        """
        Tests happy path for get_token, which also logs in
        """

        mockPost.return_value = mocked_requests_post()
        mockGetPass.return_value = "password"
        token = client.get_token()
        self.assertEqual(mockPost.return_value.status_code, 200)
        self.assertEqual(token, 'value1')
    
    @patch.object(APIClient,"post")
    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_upload_pdf(self, mockPost, mockOpen):

        """
        Tests happy path for upload_pdf
        """

        mockPost.return_value = mocked_requests_post()
        client.upload_pdf_submission('123', '1', 'email', 'file.txt')
        self.assertEqual(mockPost.return_value.status_code, 200)
    

    @patch.object(APIClient,"post")
    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_replace_pdf(self, mockPost, mockOpen):

        """
        Tests happy path for replace_pdf
        """

        mockPost.return_value = mocked_requests_post()
        client.replace_pdf_submission('123', '1', 'email', 'file.txt')
        self.assertEqual(mockPost.return_value.status_code, 200)
    
    @patch.object(APIClient,"post")
    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_upload_programming(self, mockPost, mockOpen):

        """
        Tests happy path for uploading programming solns
        """

        mockPost.return_value = mocked_requests_post()
        client.upload_programming_submission('123', '1', 'email', ['python.py', 'nb.ipynb'])
        self.assertEqual(mockPost.return_value.status_code, 200)




import unittest
import sys

if len(sys.argv) == 1:
    module = None
    test_name = None

else:
    module = "test"
    test_name = None

    # extract top-level prog
    if len(sys.argv) >= 2:
        module += f".test_{sys.argv[1]}"

    # extract next prog if applicable
    if len(sys.argv) >= 3:
        if sys.argv[1] in ["generate", "service", "execute"]:
            module += f".test_{sys.argv[2]}"
        else:
            test_name = f"Test{sys.argv[1].capitalize()}.test_{sys.argv[2]}"

    # extract test name
    if len(sys.argv) >= 4:
        test_name = f"Test{sys.argv[2].capitalize()}.test_{sys.argv[3]}"

if __name__ == "__main__":
    unittest.main(module=module, argv=[__file__], defaultTest=test_name, verbosity=0)



"""
Metadata parsers for Otter Grade
"""

import yaml
import json
import glob
import os
import re
import shutil

from .utils import list_files


FILENAME_REGEX = r"^.+\."


class GradescopeParser:
	"""Metadata parser for Gradescope exports

    Args:
        submissions_dir (``str``): path to directory with student submission files which contains 
			the ``submission_metadata.yml`` file
	"""

	def __init__(self, submissions_dir):
		# open metadata file and load into Python object
		with open(os.path.join(submissions_dir, "submission_metadata.yml")) as f:
			metadata = yaml.safe_load(f)

		# initialize metadata list
		self._metadata = []
		for folder in metadata:
			# copy submission contents into export root directory
			for file in os.listdir(os.path.join(submissions_dir, folder)):
				if os.path.isfile(os.path.join(submissions_dir, folder, file)) \
					and file[-6:] == ".ipynb":
					# create new filename
					new_filename = re.sub(FILENAME_REGEX, folder + ".", file)

					# ensure we are not overwriting another IPYNB file
					assert not os.path.exists(os.path.join(submissions_dir, folder + ".ipynb")), \
					"Extracting files from folder {} would overwrite another file in the\
					submissions directory.".format(folder)

					# copy the file
					shutil.copy(os.path.join(submissions_dir, folder, file), 
						os.path.join(submissions_dir, new_filename))

			# iterate through submitters for group submissions
			for submitter in metadata[folder][":submitters"]:
				self._metadata += [{

					# metadata is separated by filename into a list of submitters
					# and we use the name as the identifier
					"identifier": submitter[":name"],
					"filename": folder + ".ipynb"
				}]

		self._file_to_id = {file["filename"] : file["identifier"] for file in self._metadata}
		self._id_to_file = {file["identifier"] : file["filename"] for file in self._metadata}

	def get_metadata(self):
		"""Returns mapping of identifiers to files
		
		Returns:
			``list`` of ``dict``: list of dictionaries with identifier and filename information for 
				each submission
		"""
		return self._metadata

	def file_to_id(self, file):
		"""Returns a identifier given a filename
		
		Args:
			file (``str``): filename of a submission
		
		Returns:
			``str``: identifier corresponding to filename
		"""
		return self._file_to_id[file]

	def id_to_file(self, identifier):
		"""Returns a filename given an identifier
		
		Args:
			identifier (``str``): identifier of a submitter

		Returns:
			``str``: filename of that submitter's submission
		"""
		return self._id_to_file[identifier]

	def get_identifiers(self):
		"""Returns list of submission identifiers
		
		Returns:
			``list``: all identifiers from this parser
		"""
		return [file["identifier"] for file in self._metadata]

	def get_filenames(self):
		"""Returns list of filenames in the submission directory
		
		Returns:
			``list``: all filenames from this parser
		"""
		return [file["filename"] for file in self._metadata]

	def get_folders(self):
		"""Returns a list of the submission folder names from this parser

		Returns:
			``list``: the folder basenames
		"""
		return list(self._metadata.keys())


class CanvasParser:
	"""Metadata parser for Canvas exports
	
	Args:
        submissions_dir (``str``): path to directory with student submission files
	"""
	def __init__(self, submissions_dir):
		# list all files in the submissions directory since Canvas
		# doesn't output a metadata file
		submissions = list_files(submissions_dir)

		# intialize metadata list and iterate through the files
		self._metadata = []
		for file in submissions:
			# extract the identifier from the filename, which is everything
			# before the first underscore, meaning
			# somestudent_12345_67890_SomeStudent.ipynb -> somestudent
			identifier = re.match(r"(\w+?)\_", file)[1]

			# add the identifier and filename to the metadata list
			self._metadata += [{
				"identifier": identifier,
				"filename": file
			}]

		self._file_to_id = {file["filename"] : file["identifier"] for file in self._metadata}
		self._id_to_file = {file["identifier"] : file["filename"] for file in self._metadata}

	def get_metadata(self):
		"""Returns mapping of identifiers to files
		
		Returns:
			``list`` of ``dict``: list of dictionaries with identifier and filename information for 
				each submission
		"""
		return self._metadata

	def file_to_id(self, file):
		"""Returns a identifier given a filename
		
		Args:
			file (``str``): filename of a submission
		
		Returns:
			``str``: identifier corresponding to filename
		"""
		return self._file_to_id[file]

	def id_to_file(self, identifier):
		"""Returns a filename given an identifier
		
		Args:
			identifier (``str``): identifier of a submitter

		Returns:
			``str``: filename of that submitter's submission
		"""
		return self._id_to_file[identifier]

	def get_identifiers(self):
		"""Returns list of submission identifiers
		
		Returns:
			``list``: all identifiers from this parser
		"""
		return [file["identifier"] for file in self._metadata]

	def get_filenames(self):
		"""Returns list of filenames in the submission directory
		
		Returns:
			``list``: all filenames from this parser
		"""
		return [file["filename"] for file in self._metadata]


class JSONParser:
	"""Metadata parser for JSON format
	
	Args:
        file_path (``str``): path to JSON metadata file
	
	"""
	def __init__(self, file_path):
		# open JSON file and load contents into a string
		with open(file_path) as f:
			metdata = f.read()

		# parse string with json library
		self._metadata = json.loads(metdata, parse_int=str)

		# check that the JSON file has the correct structure
		assert type(self._metadata) == list, "JSON metadata is not a list"
		
		# check that each item in the JSON file is of correct type
		# and has "identifier" and "filename" keys
		for file in self._metadata:
			assert type(file) == dict, "JSON metadata does not have dictionaries"
			assert "identifier" in file.keys(), "JSON metadata does not contain \"identifier\" key"
			assert "filename" in file.keys(), "JSON metadata does not contain \"filename\" key"

		self._file_to_id = {file["filename"] : file["identifier"] for file in self._metadata}
		self._id_to_file = {file["identifier"] : file["filename"] for file in self._metadata}

	def get_metadata(self):
		"""Returns mapping of identifiers to files
		
		Returns:
			``list`` of ``dict``: list of dictionaries with identifier and filename information for 
				each submission
		"""
		return self._metadata

	def file_to_id(self, file):
		"""Returns a identifier given a filename
		
		Args:
			file (``str``): filename of a submission
		
		Returns:
			``str``: identifier corresponding to filename
		"""
		return self._file_to_id[file]

	def id_to_file(self, identifier):
		"""Returns a filename given an identifier
		
		Args:
			identifier (``str``): identifier of a submitter

		Returns:
			``str``: filename of that submitter's submission
		"""
		return self._id_to_file[identifier]

	def get_identifiers(self):
		"""Returns list of submission identifiers
		
		Returns:
			``list``: all identifiers from this parser
		"""
		return [file["identifier"] for file in self._metadata]

	def get_filenames(self):
		"""Returns list of filenames in the submission directory
		
		Returns:
			``list``: all filenames from this parser
		"""
		return [file["filename"] for file in self._metadata]


class YAMLParser:
	"""Metadata parser for YAML format
	
	Args:
        file_path (``str``): path to YAML metadata file
	"""
	def __init__(self, file_path):
		# open the YAML file and parse with yaml library
		with open(file_path) as f:
			self._metadata = yaml.safe_load(f)

		# convert any integer identifiers to strings
		for file in self._metadata:
			file["identifier"] = str(file["identifier"])

		# check that the YAML file has the correct structure
		assert type(self._metadata) == list, "YAML metadata is not a list"
		
		# check that each item in the YAML file is of correct type
		# and as "identifier" and "filename" keys
		for file in self._metadata:
			assert type(file) == dict, "YAML metadata does not have dictionaries"
			assert "identifier" in file.keys(), "YAML metadata does not contain \"identifier\" key"
			assert "filename" in file.keys(), "YAML metadata does not contain \"filename\" key"

		self._file_to_id = {file["filename"] : file["identifier"] for file in self._metadata}
		self._id_to_file = {file["identifier"] : file["filename"] for file in self._metadata}

	def get_metadata(self):
		"""Returns mapping of identifiers to files
		
		Returns:
			``list`` of ``dict``: list of dictionaries with identifier and filename information for 
				each submission
		"""
		return self._metadata

	def file_to_id(self, file):
		"""Returns a identifier given a filename
		
		Args:
			file (``str``): filename of a submission
		
		Returns:
			``str``: identifier corresponding to filename
		"""
		return self._file_to_id[file]

	def id_to_file(self, identifier):
		"""Returns a filename given an identifier
		
		Args:
			identifier (``str``): identifier of a submitter

		Returns:
			``str``: filename of that submitter's submission
		"""
		return self._id_to_file[identifier]

	def get_identifiers(self):
		"""Returns list of submission identifiers
		
		Returns:
			``list``: all identifiers from this parser
		"""
		return [file["identifier"] for file in self._metadata]

	def get_filenames(self):
		"""Returns list of filenames in the submission directory
		
		Returns:
			``list``: all filenames from this parser
		"""
		return [file["filename"] for file in self._metadata]



##################################
##### Tests for otter assign #####
##################################

import unittest
import sys
import os
import shutil
import subprocess

from subprocess import PIPE
from glob import glob
from unittest.mock import patch

from otter.argparser import get_parser
from otter.assign import main as assign
from otter.generate.token import APIClient

from . import TestCase

parser = get_parser()

TEST_FILES_PATH = "test/test-assign/"

class TestAssign(TestCase):

    def check_gradescope_zipfile(self, path, correct_dir_path):
        # unzip the zipfile
        unzip_command = ["unzip", "-o", path, "-d", TEST_FILES_PATH + "autograder"]
        unzip = subprocess.run(unzip_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(len(unzip.stderr), 0, unzip.stderr.decode("utf-8"))

        self.assertDirsEqual(TEST_FILES_PATH + "autograder", correct_dir_path, ignore_ext=[])

        # cleanup
        if os.path.exists(TEST_FILES_PATH + "autograder"):
            shutil.rmtree(TEST_FILES_PATH + "autograder")

    def test_convert_example(self):
        """
        Checks that otter assign filters and outputs correctly
        """
        # run otter assign
        run_assign_args = [
            "assign", "--no-run-tests", TEST_FILES_PATH + "example.ipynb", TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_assign_args)
        args.func = assign
        args.func(args)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "example-correct")

    def test_otter_example(self):
        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .otter file
        """
        run_assign_args = [
            "assign", "--no-init-cell", "--no-check-all", TEST_FILES_PATH + "generate-otter.ipynb", 
            TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_assign_args)
        args.func = assign
        args.func(args)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "otter-correct")     

    def test_pdf_example(self):
        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file along with PDFs
        """
        run_assign_args = [
            "assign", "--no-run-tests", TEST_FILES_PATH + "generate-pdf.ipynb", TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_assign_args)
        args.func = assign
        args.func(args)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "pdf-correct", ignore_ext=[".pdf",".zip"])
        
        # # check gradescope zip file
        # self.check_gradescope_zipfile(
        #     TEST_FILES_PATH + "output/autograder/autograder.zip", TEST_FILES_PATH + "pdf-autograder-correct",
        # )
    
    @patch.object(APIClient,"get_token")
    def test_gradescope_example(self, mocked_client):

        """
        Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file along with PDFs.
        Additionally, includes testing Gradescope integration.
        """

        mocked_client.return_value = 'token'
        
        run_gradescope_args = [
            "assign", "--no-run-tests", "--no-check-all", TEST_FILES_PATH + "generate-gradescope.ipynb", 
            TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_gradescope_args)
        args.func = assign
        args.func(args)
        
        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "gs-correct", ignore_ext=[".pdf",".zip"])

        # check gradescope zip file
        self.check_gradescope_zipfile(
            TEST_FILES_PATH + "output/autograder/autograder.zip", TEST_FILES_PATH + "gs-autograder-correct",
        )

    def test_r_example(self):
        """
        Checks that otter assign works for R notebooks correctly
        """
        run_assign_args = [
            "assign", TEST_FILES_PATH + "r-example.ipynb", 
            TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_assign_args)
        args.func = assign
        args.func(args)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "r-correct", ignore_ext=[".pdf",".zip"])
        
        # # check gradescope zip file
        # self.check_gradescope_zipfile(
        #     TEST_FILES_PATH + "output/autograder/autograder.zip", TEST_FILES_PATH + "r-autograder-correct",
        # )

    def test_rmd_example(self):
        """
        Checks that otter assign works for Rmd files
        """
        run_assign_args = [
            "assign", TEST_FILES_PATH + "rmd-example.Rmd", 
            TEST_FILES_PATH + "output", "data.csv"
        ]
        args = parser.parse_args(run_assign_args)
        args.func = assign
        args.func(args)

        self.assertDirsEqual(TEST_FILES_PATH + "output", TEST_FILES_PATH + "rmd-correct", ignore_ext=[".zip"])
        
        # check gradescope zip file
        self.check_gradescope_zipfile(
            TEST_FILES_PATH + "output/autograder/autograder.zip", TEST_FILES_PATH + "rmd-autograder-correct",
        )

    def tearDown(self):
        # cleanup
        if os.path.exists(TEST_FILES_PATH + "output"):
            shutil.rmtree(TEST_FILES_PATH + "output")
        



##################################
##### Tests for otter notebook #####
##################################

import unittest
import sys
import os
import shutil
import subprocess

from subprocess import PIPE
from glob import glob
from unittest.mock import patch

from otter import Notebook
from otter.check.logs import LogEntry, EventType, Log
from otter.check.notebook import _OTTER_LOG_FILENAME

from . import TestCase

TEST_FILES_PATH = "test/test-notebook/"

# functions used in one of the tests below
def square(x):
    return x**2

def negate(x):
    return not x

class TestNotebook(TestCase):
    """
    Test cases for the ``Notebook`` class
    """

    def test_check(self):
        """
        Checks that the otter.Notebook class works correctly
        """
        grader = Notebook(TEST_FILES_PATH + "tests")

        def square(x):
            return x**2

        def negate(x):
            return not x

        global_env = {
            "square" : square,
            "negate" : negate
        }

        for q_path in glob(TEST_FILES_PATH + "tests/*.py"):
            q = os.path.split(q_path)[1][:-3]
            result = grader.check(q, global_env=global_env)
            if q != "q2":
                self.assertEqual(result.grade, 1, "Test {} failed".format(q))
            else:
                self.assertEqual(result.grade, 0, "Test {} passed".format(q))

    def test_check_no_env(self):
        """
        Checks that the ``Notebook`` class works correctly, with no global env input
        """
        grader = Notebook(TEST_FILES_PATH + "tests")

        for q_path in glob(TEST_FILES_PATH + "tests/*.py"):
            q = os.path.split(q_path)[1][:-3]
            result = grader.check(q) #global_env=global_env)
            if q != "q2":
                self.assertEqual(result.grade, 1, f"Test {q} expected to pass but failed:\n{result}")
            else:
                self.assertEqual(result.grade, 0, f"Test {q} expected to fail but passed:\n{result}")
    
    def test_check_all_repr(self):
        """
        Checks that the representation of results as strings is correct
        """
        grader = Notebook(TEST_FILES_PATH + "tests")

        output = str(grader.check_all())

        # checks each question substring 
        output_lst = [
            'q1:\n\n    All tests passed!\n',
            'q2:\n\n    \n    0 of 1 tests passed\n',
            'q3:\n\n    All tests passed!\n',
            'q4:\n\n    All tests passed!\n',
            'q5:\n\n    All tests passed!\n'
        ]

        for result in output_lst:
            self.assertTrue(output.count(result) == 1)

    def test_export(self):
        """
        Checks export contents for existence of PDF and equality of zip
        """
        grader = Notebook(TEST_FILES_PATH + "tests")
        grader.export(TEST_FILES_PATH + "test-nb.ipynb", filtering=False)

        self.assertTrue(os.path.isfile(TEST_FILES_PATH + "test-nb.pdf"))
        
        with self.unzip_to_temp(TEST_FILES_PATH + "test-nb.zip") as unzipped_dir:
            # breakpoint()
            self.assertDirsEqual(
                unzipped_dir, 
                TEST_FILES_PATH + "export-correct", 
                ignore_ext=[".pdf", ""] # second ignores .OTTER_LOG files
            )

        # cleanup
        os.remove(TEST_FILES_PATH + "test-nb.pdf")
        os.remove(TEST_FILES_PATH + "test-nb.zip")

    @patch.object(LogEntry, "shelve")
    def test_nb_log(self, mock_log):
        """
        Checks existence of log when running nb
        """

        mock_log.return_value = LogEntry(EventType.CHECK)
        grader = Notebook(TEST_FILES_PATH + "tests")
        output = grader.check_all()

        self.assertTrue(os.path.isfile(_OTTER_LOG_FILENAME))
    
    def tearDown(self):
        if os.path.isfile(_OTTER_LOG_FILENAME):
            os.remove(_OTTER_LOG_FILENAME)



##################################
##### Tests for Otter Export #####
##################################

# NOTES:
# - tests do not check for PDF equality but rather have exporter save the TeX file as well
#   and check that the TeX is correct (assuming that TeX conversion works correctly) and checks
#   that PDF file exists

import os
import unittest
import subprocess
import json
import re
import contextlib
import nbconvert
import filecmp
import nbformat

from io import StringIO
from unittest import mock
from subprocess import Popen, PIPE
from glob import glob
from textwrap import dedent

from otter.argparser import get_parser
from otter.export import export_notebook
from otter.export import main as export
from otter.export.exporters.base_exporter import BaseExporter

from . import TestCase

parser = get_parser()

TEST_FILES_PATH = "test/test-export/"

class TestExport(TestCase):

    def test_success_HTML(self):
        """
        Tests a successful export with filtering and no pagebreaks
        """
        test_file = "successful-html-test"
        grade_command = ["export", "--filtering", "-s", "-e", "latex",
            TEST_FILES_PATH + test_file + ".ipynb"
        ]
        args = parser.parse_args(grade_command)
        args.func = export
        args.func(args)

        # check existence of pdf and tex
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".pdf"))
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".tex"))

        # cleanup
        cleanup_command = ["rm", TEST_FILES_PATH + test_file + ".pdf", TEST_FILES_PATH + test_file + ".tex"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(cleanup.returncode, 0,"Error in cleanup: " + str(cleanup.stderr))
        
    def test_success_pagebreak(self):
        """
        Tests a successful filter with pagebreaks
        """
        test_file = "success-pagebreak-test"
        grade_command = ["export", "--filtering", "-e", "latex",
            "--pagebreaks", "-s",
            TEST_FILES_PATH + test_file + ".ipynb"
        ]
        args = parser.parse_args(grade_command)
        args.func = export
        args.func(args)

        # check existence of pdf and tex
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".pdf"))
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".tex"))

        # cleanup
        cleanup_command = ["rm", TEST_FILES_PATH + test_file + ".pdf", TEST_FILES_PATH + test_file + ".tex"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(cleanup.returncode, 0,"Error in cleanup: " + str(cleanup.stderr))

    def test_no_open(self):
        """
        Tests a filtered export without an opening comment
        """
        test_file = "no-open-tag-test"
        grade_command = ["export", "--filtering", "-e", "latex",
            TEST_FILES_PATH + test_file + ".ipynb"
        ]

        should_contain = [
            "There was an error generating your LaTeX\nShowing concise error message\n" + "=" * 60,
            "(There were 3 error messages)",
            "This is BibTeX"
        ]

        args = parser.parse_args(grade_command)
        args.func = export

        actual_output = StringIO()
        with contextlib.redirect_stdout(actual_output):
            args.func(args)

        for s in should_contain:
            self.assertIn(s, actual_output.getvalue(), 
                f"Empty Tex did not contain substring: {s}\n\n"
                f"Output:\n\n{actual_output.getvalue()}")

    def test_no_close(self):
        """
        Tests a filtered export without a closing comment
        """
        test_file = "no-close-tag-test"
        grade_command = ["export", "--filtering", "-e", "latex",
            "--pagebreaks", "-s",
            TEST_FILES_PATH + test_file + ".ipynb"
        ]
        args = parser.parse_args(grade_command)
        args.func = export
        args.func(args)

        # check existence of pdf and tex
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".pdf"))
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".tex"))

        # cleanup

        cleanup_command = ["rm", TEST_FILES_PATH + test_file + ".pdf", TEST_FILES_PATH + test_file + ".tex"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(cleanup.returncode, 0,"Error in cleanup:" + str(cleanup.stderr))

    def test_load_notebook(self):
        """
        Tests a successful load_notebook
        """
        test_file = "successful-html-test"
        node = BaseExporter.load_notebook(TEST_FILES_PATH + test_file + ".ipynb", filtering=True)

        nbformat.write(node, TEST_FILES_PATH + test_file)

        # check existence of file
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file))
        self.assertTrue(filecmp.cmp(TEST_FILES_PATH + test_file, TEST_FILES_PATH + "correct/" + test_file))
        
        # cleanup
        cleanup_command = ["rm", TEST_FILES_PATH + test_file]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(cleanup.returncode, 0,"Error in cleanup: " + str(cleanup.stderr))



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

# get version
env = {}
with open("otter/version.py") as f:
	exec(f.read(), env)
version = env["__version__"]

setuptools.setup(
	name = "otter-grader",
	version = version,
	author = "Chris Pyles",
	author_email = "cpyles@berkeley.edu",
	description = "Python and Jupyter Notebook autograder",
	long_description = long_description,
	long_description_content_type = "text/markdown",
	url = "https://github.com/ucbds-infra/otter-grader",
	license = "BSD-3-Clause",
	packages = setuptools.find_packages(exclude=["test"]),
	classifiers = [
		"Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
	],
	install_requires=[
		"pyyaml", "nbformat", "ipython", "nbconvert<6.0.0", "tqdm", "setuptools", "pandas", "tornado",
		"docker", "jinja2", "dill", "pdfkit", "PyPDF2"
	],
	scripts=["bin/otter"],
	package_data={
		"otter.service": ["templates/*.html"], 
		"otter.export.exporters": ["templates/*"],
		"otter.generate": ["templates/*"],
	},
)



"""
ottr test adapters for Otter Assign
"""

import re
import pprint
import yaml
import nbformat

from collections import namedtuple

from ..constants import TEST_REGEX, OTTR_TEST_NAME_REGEX, OTTR_TEST_FILE_TEMPLATE
from ..tests import write_test
from ..utils import get_source, lock

Test = namedtuple('Test', ['name', 'hidden', 'body'])

def read_test(cell, question, assignment, rmd=False):
    """
    Returns the contents of a test as an ``(input, output, hidden)`` named tuple
    
    Args:
        cell (``nbformat.NotebookNode``): a test cell
        question (``dict``): question metadata
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        rmd (``bool``, optional): whether the cell is from an Rmd file; if true, the first and last
            lines of ``cell``'s source are trimmed, since they should be backtick delimeters

    Returns:
        ``Test`` or ``OttrTest``: test named tuple
    """
    if rmd:
        source = get_source(cell)[1:-1]
    else:
        source = get_source(cell)
    hidden = bool(re.search("hidden", source[0], flags=re.IGNORECASE))
    lines = source[1:]
    assert sum("test_that(" in line for line in lines) == 1, \
        f"Too many test_that calls in test cell (max 1 allowed):\n{cell}"
    test_name = None
    for line in lines:
        match = re.match(OTTR_TEST_NAME_REGEX, line)
        if match:
            test_name = match.group(1)
            break
    assert test_name is not None, f"Could not parse test name:\n{cell}"
    return Test(test_name, hidden, '\n'.join(lines))

def gen_test_cell(question, tests, tests_dict, assignment):
    """
    Parses a list of test named tuples and creates a single test file. Adds this test file as a value
    to ``tests_dict`` with a key corresponding to the test's name, taken from ``question``. Returns
    a code cell that runs the check on this test.
    
    Args:
        question (``dict``): question metadata
        tests (``list`` of ``Test``): tests to be written
        tests_dict (``dict``): the tests for this assignment
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``nbformat.NotebookNode``: code cell calling ``otter.Notebook.check`` on this test
    """
    cell = nbformat.v4.new_code_cell()
    cell.source = ['. = ottr::check("tests/{}.R")'.format(question['name'])]

    points = question.get('points', len(tests))
    if isinstance(points, int):
        if points % len(tests) == 0:
            points = [points // len(tests) for _ in range(len(tests))]
        else:
            points = [points / len(tests) for _ in range(len(tests))]
    assert isinstance(points, list) and len(points) == len(tests), \
        f"Points for question {question['name']} could not be parsed:\n{points}"

    test = gen_suite(question['name'], tests, points)

    tests_dict[question['name']] = test
    lock(cell)
    return cell

def gen_suite(name, tests, points):
    metadata = {'name': name, 'cases': []}
    cases = metadata['cases']
    for test, p in zip(tests, points):
        cases.append({
            'name': test.name,
            'points': p,
            'hidden': test.hidden
        })
    
    metadata = yaml.dump(metadata)

    return OTTR_TEST_FILE_TEMPLATE.render(
        metadata = metadata,
        tests = tests
    )

def remove_hidden_tests_from_dir(test_dir, assignment):
    """
    Rewrites test files in a directory to remove hidden tests
    
    Args:
        test_dir (``pathlib.Path``): path to test files directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
    """
    for f in test_dir.iterdir():
        if f.suffix != '.R':
            continue

        with open(f) as f2:
            test = f2.read()
        
        metadata, in_metadata, start_lines, test_names = "", False, {}, []
        metadata_start, metadata_end = -1, -1
        lines = test.split("\n")
        for i, line in enumerate(lines):
            match = re.match(OTTR_TEST_NAME_REGEX, line)
            if line.strip() == "test_metadata = \"":
                in_metadata = True
                metadata_start = i
            elif in_metadata and line.strip() == "\"":
                in_metadata = False
                metadata_end = i
            elif in_metadata:
                metadata += line + "\n"
            elif match:
                test_name = match.group(1)
                test_names.append(test_name)
                start_lines[test_name] = i

        assert metadata and metadata_start != -1 and metadata_end != -1, \
            f"Failed to parse test metadata in {f}"
        metadata = yaml.full_load(metadata)
        cases = metadata['cases']

        lines_to_remove, cases_to_remove = [], []
        for i, case in enumerate(cases):
            if case['hidden']:
                start_line = start_lines[case['name']]
                try:
                    next_test = test_names[test_names.index(case['name']) + 1]
                    end_line = start_lines[next_test]
                except IndexError:
                    end_line = len(lines)
                lines_to_remove.extend(range(start_line, end_line))
                cases_to_remove.append(i)

        metadata['cases'] = [c for i, c in enumerate(cases) if i not in set(cases_to_remove)]
        lines = [l for i, l in enumerate(lines) if i not in set(lines_to_remove)]
        lines[metadata_start:metadata_end + 1] = ["test_metadata = \""] + \
            yaml.dump(metadata).split("\n") + ["\""]
        test = "\n".join(lines)

        write_test(f, test)
